
#include <stdio.h>
#include <unistd.h>

#include <pthread.h>

#include <sched.h>

#include <glibtop/proclist.h>
#include <glibtop/xmalloc.h>


static int increasetime=1; // Seconds between each time the SCHED_OTHER thread is increasing the counter.
static int checktime=4; // Seconds between each time the SCHED_FIFO thread checks that the counter is increased.
static int waittime=8;  // Seconds the SCHED_FIFO thread waits before setting the processes back to SCHED_FIFO.

struct das_proclist{
  pid_t pid;
  int policy; //SCHED_OTHER, SCHED_FIFO, SHED_RR
  int priority;
};


static int counter=0;

// The SCHED_OTHER thread.
static void *counter_func(void *arg){

  {
    // Eh, this is not necessary. But, well, just in case.
    struct sched_param par;
    par.sched_priority = 0;
    if(sched_setscheduler(0,SCHED_OTHER,&par)==-1){
      fprintf(stderr,"das_watchdog: Unable to set SCHED_other...\n");
    }
  }

  for(;;){
    counter++;
    sleep(increasetime);
  }
  return NULL;
}


struct das_proclist *get_proclist(int *num_procs){
  int lokke=0;
  glibtop_proclist proclist_def={0};
  unsigned *proclist=glibtop_get_proclist(&proclist_def,GLIBTOP_KERN_PROC_ALL,0); //|GLIBTOP_EXCLUDE_SYSTEM,0);
  struct das_proclist *ret=calloc(sizeof(struct das_proclist),proclist_def.number);

  *num_procs=proclist_def.number;

  for(lokke=0;lokke<proclist_def.number;lokke++){
    struct sched_param par;
    ret[lokke].pid=proclist[lokke];
    ret[lokke].policy=sched_getscheduler(proclist[lokke]);
    sched_getparam(proclist[lokke],&par);
    ret[lokke].priority=par.sched_priority;
    //printf("%u,",proclist[lokke]);
  }
  printf("\n\n%d %d %d\n",(int)proclist_def.number,(int)proclist_def.total,(int)proclist_def.size);

  glibtop_free(proclist);

  return ret;
}





// The SCHED_FIFO thread.
int main(int argc,char **argv){
  pid_t mypid=getpid();
  pthread_t counter_thread={0};

  pthread_create(&counter_thread,NULL,counter_func,NULL);

  if(getuid()!=0){
    printf("Warning, you are not running as root. das_watchdog should be run as an init-script at startup, and not as a normal user.\n");
  }

  if(access(WHICH_WISH,X_OK)!=0){
    printf("Warning, \"wish\" is not found or is not an executable. You will not receive messages to the screen in case das_watchdog has to take action.\n");
  }


  {
    struct sched_param par;
    par.sched_priority = sched_get_priority_max(SCHED_FIFO);
    if(sched_setscheduler(0,SCHED_FIFO,&par)==-1){
      fprintf(stderr,"das_watchdog: Unable to set SCHED_FIFO realtime priority for the watchdog thread. Exiting.\n");
      return 0;
    }
  }

  for(;;){
    int lastcounter=counter;
    sleep(checktime);
    //printf("counter: %d\n",counter);
    if(lastcounter==counter){ // || counter==2){
      int num_procs=0;
      struct das_proclist *proclist=get_proclist(&num_procs);
      int lokke;
      printf("Das_Watchdog: Watchdog kicking!\n");

      for(lokke=0;lokke<num_procs;lokke++){
	if(proclist[lokke].policy!=SCHED_OTHER && proclist[lokke].pid!=mypid){
	  struct sched_param par={0};
	  par.sched_priority=0;
	  sched_setscheduler(proclist[lokke].pid,SCHED_OTHER,&par);	  
	  printf("Das_Watchdog: Setting pid %d temporarily to SCHED_FIFO.\n",proclist[lokke].pid);
	}
      }
      {
	char temp[5000];
	//sprintf(temp,"%s -geometry 600x50 -name \"WARNING! das_watchdog pauses realtime operations for %d seconds.\"&",WHICH_WISH,waittime);
	sprintf(temp,"xmessage \"WARNING! das_watchdog pauses realtime operations for %d seconds.\"&",waittime);

	//seteuid(500);
	system("xhost +");
	setenv("DISPLAY",":0.0",1);
	//seteuid(0);
	//system("su -l root -c \"xmessage asd\"");
	//system("xmessage asd");


	system(temp);
	//system("su root -c \"xmessage wef\"");
	//system("ssh -X 127.0.0.1 /usr/X11R6/bin/xmessage gakk");
	//seteuid(0);
      }
      sleep(waittime);
      for(lokke=0;lokke<num_procs;lokke++){
	if(proclist[lokke].policy!=SCHED_OTHER && proclist[lokke].pid!=mypid){
	  struct sched_param par={0};
	  par.sched_priority=proclist[lokke].priority;
	  sched_setscheduler(proclist[lokke].pid,proclist[lokke].policy,&par);	  
	  printf("Das_Watchdog: Setting pid %d back to realtime priority.\n",proclist[lokke].pid);
	}
      }
      free(proclist);
    }
  }

  return 0;
}
